
import numpy as np
from sklearn.tree import DecisionTreeRegressor

"""
We implement sklearn.tree.DecisionTreeRegressor package to construct a single tree.

_get_init_val(): Initialize the model with a constant (Hill estimator);
_get_gradient(): Compute the current negative gradient;
_update_score(): Update leaf values of each regression tree;

fit(): Fit a model using given data and parameters.
x: Covariates, a n*p matrix;
y: The response, a 1d array;
q: Tail fraction. The threshold wn is then denoted as the (100-q)th sample quantile of y; 
n_estimators: Number of trees;
learning_rate: Learning rate of gradient boosting;
max_depth: Maximum depth of a single tree;
min_samples_split: Stop splitting if the current samples are less than min_samples_split, default to 2;
min_samples_leaf: Any terminal node must contains at least min_samples_leaf samples, default to 2.

predict(): Predict gamma of given x0;
learn_curve(): Return the loss on {x,y} in each iteration of gradient boosting;
feature_importance(): Return the original importance measure of x_1...x_p.
 
Note that the methods of computing adjusted feature importance and partial dependence are given in DataAnalysis.py.
"""

class EVIboost():
    
    def __init__(self):
        
        self.trees = None
        self.learning_rate = None
        self.init_val = None
    
    def _get_init_val(self,y,wn):
        
        return np.sum(np.log(y/wn)*(y>wn))/np.sum(y>wn)
    
    
    def _get_gradient(self,y,yhat,wn):
        
        return ((np.log(y/wn)-yhat) / (yhat**2))*(y>wn)
    
    def _update_score(self,tree,x1,y1,wn,yhat1,gd1):
        
        t = tree.tree_
        locs = t.apply(x1.astype(np.float32))
        nx = x1.shape[0]
        
        for i in range(t.node_count):
            idx = [j for j in range(nx) if locs[j]==i]
            if len(idx) >= 1:
                val = np.sum(gd1[idx]) / np.sum(2*gd1[idx]/yhat1[idx] + 1/(yhat1[idx]**2))
                if not np.isnan(val):
                    t.value[i,0,0] = val

    def fit(self,x,y,q,n_estimators,learning_rate,max_depth,min_samples_split=2,min_samples_leaf=2):
        
        wn = np.percentile(y,100-q)
        n = y.shape[0]
        self.init_val = self._get_init_val(y,wn)
        yhat = np.full(n,self.init_val) 
        gd = self._get_gradient(y,yhat,wn)
        
        self.trees = []
        self.learning_rate = learning_rate
        
        for m in range(n_estimators):
            
            tree = DecisionTreeRegressor(max_depth=max_depth,min_samples_split=min_samples_split,min_samples_leaf=min_samples_leaf)
            tree.fit(x[y>wn],gd[y>wn])
            
            self._update_score(tree,x[y>wn],y[y>wn],wn,yhat[y>wn],gd[y>wn])
            
            yhat += learning_rate*tree.predict(x)
            gd = self._get_gradient(y,yhat,wn)
            
            self.trees.append(tree)
            
    def predict(self,x0):
        
        nx = x0.shape[0]
        gamma_hat = np.full(nx,self.init_val)
        for t in self.trees:
            gamma_hat += self.learning_rate*t.predict(x0)
        return gamma_hat
    
    def learn_curve(self,x,y,wn):
        nx = x.shape[0]
        ls = []
        yhat = np.full(nx,self.init_val)
        loss = np.sum((np.log(y/wn)/yhat+np.log(yhat))*(y>wn))
        ls.append(loss)
    
        for t in self.trees:
            yhat += self.learning_rate*t.predict(x)
            loss = np.sum((np.log(y/wn)/yhat+np.log(yhat))*(y>wn))
            ls.append(loss)
        ls = np.asarray(ls)
        return ls

    def feature_importance(self,p):
        imp = np.full(shape=p,fill_value=0.0)
        for t in self.trees:
            imp += t.feature_importances_
        return imp
            

"""
Use K-fold cross validation to select the optimal tuning parameters {M,v,L};
We tune D (maximum depth of a tree) instead of L (number of terminal nodes) for the convenience of coding (L can be determined once D is given).
"""

import matplotlib.pyplot as plt
from sklearn.model_selection import KFold

def cv(x,y,q_opt,K=5):

    train_seq,test_seq = [],[]
    kf = KFold(n_splits=K,shuffle=True)
    
    for train,test in kf.split(x):
       train_seq.append(train)
       test_seq.append(test)
    
    cv_record = []     
    maxd,max_est,vseq = 1,1000,[0.01,0.0075,0.005,0.0025]
    
    for v in vseq:
        ls = np.full(shape=max_est+1,fill_value=0.0)
        for i in range(K):
            xtr,ytr = x[train_seq[i],:],y[train_seq[i]]
            xte,yte = x[test_seq[i],:],y[test_seq[i]]
    
            model = EVIboost()
            model.fit(xtr,ytr,q_opt,n_estimators=max_est,learning_rate=v,max_depth=maxd)
            wn = np.percentile(ytr,100-q_opt)
            ls = ls+model.learn_curve(xte,yte,wn)
    
        cv_record.append(ls)
    
    plt.figure()
    for i in range(len(vseq)):
        plt.plot(cv_record[i],label='v='+str(vseq[i]))
    plt.legend()
    plt.show()
    
    for i in range(len(vseq)):
        print('\nv=' + str(vseq[i]) + ', minimum loss=' + str(min(cv_record[i])))
        print('argmin=' + str(np.argsort(cv_record[i])[0]))
