
import numpy as np, matplotlib.pyplot as plt, pandas as pd, time
from scipy.stats import norm
from tir import TIR,selectq
from eviboost import *

"""
This is the first part of simulation. We compare the MSE of EVIboost with TIR under q = 0.10,0.05,0.025 and q*.
sim(): The function to generate random samples from five cases.
"""

m,p = 0.1,10
theta1 = np.asarray([1/2,-1/3,1/3,0,0,0,0,0,0,0]).reshape((p))
theta2 = np.asarray([1/5,-1/5,1/5,-1/5,1/5,-1/5,1/5,-1/5,1/5,-1/5]).reshape((p))

mean = np.zeros(shape=(p))
cov = np.zeros(shape=(p,p))

for i in range(p):
    for j in range(p):
        cov[i,j] = 1/2**(np.abs(i-j))

def sim(n,case=1):  
    z = np.random.multivariate_normal(mean=mean,cov=cov,size=(n))
    x = (norm.cdf(z)-0.5)*np.sqrt(12)   
    a1 = 3*np.exp(x.dot(theta1)).reshape(n)  
    a2 = 3*np.exp(x.dot(theta2)).reshape(n) 
    a3 = 3*np.exp((x**2).dot(theta1))
    a4 = np.exp( (x[:,0]+x[:,1])**2  + (x[:,1]+x[:,2])**4 ).reshape(n)
    a5 = np.exp( ((x[:,0]-x[:,1])**2)**0.25 + ((x[:,2]-x[:,3])**2)**(-0.25) ).reshape(n)  
    A = [a1,a2,a3,a4,a5]
    a = A[case-1]
    u = np.random.uniform(0,1,n)  
    y = ((1+m*u)/(1-u))**(1/a)   
    return a,x,y


"""
q* = argmin(D1+D2+D3) refers to the optimal tail fraction for TIR model.
We do min-max normalization on D1,D2,D3 before summing them up.
Use the function selectq() (see tir.py) to determine q*.
"""

np.random.seed(123)
for case in [1,2,3,4,5]:
    n = 1000
    a,x,y = sim(n,case)    
    DS = selectq(x,y)
    
    (f1,f,d1) = DS[0]
    (f2,f,d2) = DS[1]
    (f3,f,d3) = DS[2]
    
    d1 = (d1-min(d1))/(max(d1)-min(d1))
    d2 = (d2-min(d2))/(max(d2)-min(d2))
    d3 = (d3-min(d3))/(max(d3)-min(d3))
    d = d1+d2+d3
    
    plt.figure()
    plt.plot(f,d1,label='D1',color='k')
    plt.plot(f,d2,label='D2',color='k',linestyle='--')
    plt.plot(f,d3,label='D3',color='blue',linestyle='-.')
    plt.xlabel('Tail fraction q, argmin(D1+D2+D3) = '+str(f[np.argsort(d)[0]]))
    plt.ylabel('Discrepancy')  
    plt.legend()
    plt.show()
    

"""
The parameters case,q,D,V,M are provided in sheet 1 of sim_setup.xlsx.
"""

n = 1000
case,q,D,V,M = ...

np.random.seed(123)
se_evis,se_tirs = [],[]
for i in range(100): 
    a,x,y = sim(n,case)
    wn = np.percentile(y,q)
    ate,xte,yte = sim(n,case)
    
    gamma_te = 1/ate
    xteh = np.hstack(( np.ones(shape=(xte.shape[0],1)),xte))
    thhat = TIR(x,y,q)
    gamma_tir = 1/np.exp(xteh.dot(thhat)).reshape(n)
    se_tir = np.mean( (gamma_tir-gamma_te)**2 ) 
    se_tirs.append(se_tir)
           
    model = EVIboost()
    model.fit(x,y,q,learning_rate=V,n_estimators=M,max_depth=D)
    gamma_evi = model.predict(xte)    
    se_evi = np.mean( (gamma_evi-gamma_te)**2 )  
    se_evis.append(se_evi)
    
    print('\n',i,'--TIR=',se_tir,'--EVI=',se_evi,'q=',q)


se = pd.DataFrame({'TIR': se_tirs, 'EVI': se_evis,'q':q})
eff = np.median(se['TIR'])/np.median(se['EVI'])
print(se.describe())

plt.figure()
se[['TIR','EVI']].plot.box()
plt.title('MSE',fontsize=14)
plt.xlabel('eff= '+str(eff))
plt.show()


"""
This is the second part of simulation. 
We set two values for m (0.10,15) and compare the MSE of EVIboost with TIR under each q of 5,15,25...95.
The parameters case,q,D,V,M,m are provided in sheet2 of sim_setup.xlsx.
"""

case,q,D,V,M,m = ...

np.random.seed(123)
se_tir,se_evi = 0,0
for r in range(100):
    a,x,y = sim(n,case)
    ate,xte,yte = sim(n,case)
    gamma_te = 1/ate
    
    xteh = np.hstack((np.ones(shape=(xte.shape[0],1)),xte))
    thhat = TIR(x,y,q)
    gamma_tir = 1/np.exp(xteh.dot(thhat)).reshape(n)
    se_tir += np.mean( (gamma_tir-gamma_te)**2 ) / 100
           
    model = EVIboost()
    model.fit(x,y,q,learning_rate=V,n_estimators=M,max_depth=D)
    gamma_evi = model.predict(xte)
    se_evi += np.mean((gamma_evi-gamma_te)**2) / 100  
    
print('\n--TIR=',se_tir,'--EVI=',se_evi,'q=',q)
print('\n eff=',se_tir / se_evi)


"""
Comparison on the computational time of the two models.
"""

case,n = ...
q,D,V,M,m = 10,1,0.001,1000,0.1

Data = []
np.random.seed(123)
for r in range(100):
    a,x,y = sim(n,case)
    Data.append((x,y))

start = time.time()
for x,y in Data:
    model = EVIboost()
    model.fit(x,y,q,learning_rate=V,n_estimators=M,max_depth=D)  
t = time.time()-start

start = time.time()
for x,y in Data:
    thhat = TIR(x,y,q)
t_tir = time.time()-start

