
import numpy as np
from scipy.optimize import minimize
from eviboost import *

"""
We implment scipy.optimize.minimize to do BFGS.

TIR(x,y,q): return the estimated coefficients.
x: Covariates (without the constant), a n*p matrix;
y: The response, a 1d array;
q: Tail fraction;
LS(): The loss function under estimated coefficients.

selectq(): Compute the three discrepancy measures D1,D2,D3 under TIR (default) or EVIboost model.
We range the tail fraction q from 0.5(%) to 99.5(%) by 0.5(%). 
Return a list that contains three tuples. For each tuple (q,qlist,D):
q: argmin_{q in qlist} D;
qlist: The domain of q;
D: The discrepancy measure (D1,D2 or D3) under each q of qlist.
"""


def TIR(x,y,q):  
    xh = np.hstack((np.ones(shape=(x.shape[0],1)),x))
    wn = np.percentile(y,100-q)
    
    def LS(thhat):  
        ls = np.sum((np.exp(xh.dot(thhat)) * np.log(y/wn) - xh.dot(thhat))*(y>wn))
        return ls

    p = x.shape[1]
    th0 = np.full(p+1,0.5) 
    opt = minimize(LS,th0,method='BFGS')
    return opt.x


def selectq(x,y,model='TIR'):
    
    qlist = np.linspace(0.5,99.5,199)
    ds1,ds2,ds3 = [],[],[]
    
    for q in qlist:
        print(q)
        
        wn_q = np.percentile(y,100-q)
        
        if model == 'EVI':
            m = EVIboost()
            m.fit(x,y,q,n_estimators=100,learning_rate=0.001,max_depth=1)
            gamma_hat = m.predict(x)
            U = np.exp(-(1/gamma_hat) * np.log(y/wn_q))
        
        else:
            thhat_q = TIR(x,y,q) 
            xh = np.hstack((np.ones(shape=(x.shape[0],1)),x))
            U = np.exp(-np.exp(xh.dot(thhat_q)) * np.log(y/wn_q))
            
        U = U[y>wn_q]
        n0 = U.shape[0]
        D1,D2,D3 = 0,0,0 

        for i in range(n0):
            D1 += (U[i]-np.sum(U<=U[i])/n0)**2
        D1 = D1/n0
        
        D2 = np.abs(U[0]-np.sum(U<=U[0])/n0)
        for i in range(n0):
            D_new = np.abs(U[i]-np.sum(U<=U[i])/n0)
            if D2 <= D_new:
                D2 = D_new
                
        for i in range(n0):
            D3 += (U[i]-np.sum(U<=U[i])/n0)**2 / (U[i]*(1-U[i]))
        D3 = D3/n0
        
        ds1.append(D1)
        ds2.append(D2)
        ds3.append(D3)
    
    
    q1 = qlist[ds1.index(np.min(ds1))]
    q2 = qlist[ds2.index(np.min(ds2))]
    q3 = qlist[ds3.index(np.min(ds3))]
        
    ds1 = np.asarray(ds1)
    ds2 = np.asarray(ds2)
    ds3 = np.asarray(ds3)
   
    return [(q1,qlist,ds1),(q2,qlist,ds2),(q3,qlist,ds3)]

